package de.tudresden.inf.st.eraser;

import de.tudresden.inf.st.eraser.jastadd.model.*;
import de.tudresden.inf.st.eraser.util.TestUtils;
import de.tudresden.inf.st.eraser.util.TestUtils.ModelAndItem;
import org.junit.jupiter.api.Tag;
import org.junit.jupiter.api.Test;

import java.nio.charset.StandardCharsets;
import java.util.ArrayList;
import java.util.List;
import java.util.concurrent.TimeUnit;

import static de.tudresden.inf.st.eraser.util.TestUtils.getMqttHost;
import static org.assertj.core.api.Assertions.assertThat;
import static org.junit.jupiter.api.Assertions.assertEquals;
import static org.junit.jupiter.api.Assertions.assertTrue;

/**
 * Test for everything related to MQTT.
 *
 * @author rschoene - Initial contribution
 */
@Tag("mqtt")
public class MqttTests {

  private static final String firstPart = "a";
  private static final String alternativeFirstPart = "x";
  private static final String secondPart = "b";
  private static final double firstState = 1.0;
  private static final double secondState = 2.0;
  private static final double thirdState = 3.0;

  @Test
  public void sendSingleItem() throws Exception {
    // Given model with single item
    ModelAndItem modelAB = createModelAndSetupMqttRoot();
    String outgoingPrefix = modelAB.model.getRoot().getMqttRoot().getOutgoingPrefix();
    String expectedTopic = outgoingPrefix + firstPart + "/" + secondPart;

    MqttHandler handler = new MqttHandler().dontSendWelcomeMessage().setHost(getMqttHost());
    assertTrue(handler.waitUntilReady(2, TimeUnit.SECONDS), "Could not connect to MQTT broker");
    List<String> messages = new ArrayList<>();
    handler.newConnection(expectedTopic, bytes -> messages.add(new String(bytes)));

    NumberItem sut = modelAB.item;
    sut.setTopicString(firstPart + "/" + secondPart);
    assertTrue(modelAB.model.connectAllItems(), "Could not connect items");
    TestUtils.waitForMqtt();

    // When setting first state
    sut.setState(firstState);

    // Then a message with first state should be sent
    TestUtils.waitForMqtt();
    assertEquals(1, messages.size());
    assertEquals(Double.toString(firstState), messages.get(0));

    // When setting second state
    sut.setState(secondState);

    // Then a message with second state should be sent
    TestUtils.waitForMqtt();
    assertEquals(2, messages.size());
    assertEquals(Double.toString(secondState), messages.get(1));
  }

  @Test
  public void sendItemsDifferentTopic() throws Exception {
    // Given model with two items, each with a different topic
    ModelAndItem modelAB = createModelAndSetupMqttRoot();
    String outgoingPrefix = modelAB.model.getRoot().getMqttRoot().getOutgoingPrefix();
    String expectedTopic1 = outgoingPrefix + firstPart + "/" + secondPart;
    String expectedTopic2 = outgoingPrefix + alternativeFirstPart + "/" + secondPart;

    MqttHandler handler = new MqttHandler().dontSendWelcomeMessage().setHost(getMqttHost());
    assertTrue(handler.waitUntilReady(2, TimeUnit.SECONDS), "Could not connect to MQTT broker");
    List<String> messagesTopic1 = new ArrayList<>();
    List<String> messagesTopic2 = new ArrayList<>();
    handler.newConnection(expectedTopic1, bytes -> messagesTopic1.add(new String(bytes)));
    handler.newConnection(expectedTopic2, bytes -> messagesTopic2.add(new String(bytes)));

    NumberItem item1 = modelAB.item;
    NumberItem item2 = TestUtils.addItemTo(modelAB.model, 0, true);
    item1.setTopicString(firstPart + "/" + secondPart);
    item2.setTopicString(alternativeFirstPart + "/" + secondPart);
    assertTrue(modelAB.model.connectAllItems(), "Could not connect items");
    TestUtils.waitForMqtt();

    // When setting first state for both items
    item1.setState(firstState);
    item2.setState(firstState);

    // Then a message with first state should be sent to each topic
    TestUtils.waitForMqtt();
    assertEquals(1, messagesTopic1.size());
    assertEquals(1, messagesTopic2.size());
    assertEquals(Double.toString(firstState), messagesTopic1.get(0));
    assertEquals(Double.toString(firstState), messagesTopic2.get(0));

    // When setting second state for first item, and third state for second item
    item1.setState(secondState);
    item2.setState(thirdState);

    // Then a message with respective state should be sent to each topic
    TestUtils.waitForMqtt();
    assertEquals(2, messagesTopic1.size());
    assertEquals(2, messagesTopic2.size());
    assertThat(messagesTopic1).contains(Double.toString(secondState));
    assertThat(messagesTopic2).contains(Double.toString(thirdState));
  }

  @Test
  public void receiveSingleItem() throws Exception {
    // Given model with single item
    ModelAndItem modelAB = createModelAndSetupMqttRoot();
    String incomingPrefix = modelAB.model.getRoot().getMqttRoot().getIncomingPrefix();
    String usedTopic = incomingPrefix + firstPart + "/" + secondPart;

    MqttHandler handler = new MqttHandler().dontSendWelcomeMessage().setHost(getMqttHost());
    assertTrue(handler.waitUntilReady(2, TimeUnit.SECONDS), "Could not connect to MQTT broker");

    NumberItem sut = modelAB.item;
    sut.setTopicString(firstPart + "/" + secondPart);
    assertTrue(modelAB.model.connectAllItems(), "Could not connect items");
    TestUtils.waitForMqtt();

    // When a message is published on the topic the item is receiving
    handler.publish(usedTopic, "4.0".getBytes(StandardCharsets.UTF_8));
    TestUtils.waitForMqtt();

    // Then the state of the item should be updated
    assertEquals(4, sut.getState(), TestUtils.DELTA);
  }

  static ModelAndItem createModelAndSetupMqttRoot() {
    ModelAndItem mai = TestUtils.createModelAndItem(0, true);
    SmartHomeEntityModel model = mai.model;
    MqttRoot mqttRoot = new MqttRoot();
    mqttRoot.setIncomingPrefix("inc");
    mqttRoot.setOutgoingPrefix("out");
    mqttRoot.getHost().setHostName(getMqttHost()).setPort(1883);
    mqttRoot.ensureCorrectPrefixes();
    model.getRoot().setMqttRoot(mqttRoot);
    return mai;
  }

}
